/*
 * QFIVEDEV-381 - Разработка хранимой процедуры.
 * Добавить 3 функции:
 * 1. DATE_DIFF - Сравнение двух дат.
 * 2. GET_CLIENT_SESSIONS - Получение данных о клиентской сессии.
 * 3. REPORT_CLIENTS_SERVICE_METRICS - Отчет "Показатели обслуживания клиентов".
*/

-- Вспомогательная функция сравнения двух дат.
CREATE OR REPLACE FUNCTION DATE_DIFF(UNITS VARCHAR(30), START_T TIMESTAMP, END_T TIMESTAMP)
    RETURNS INT AS
$$
DECLARE
    DIFF_INTERVAL INTERVAL;
    DIFF          INT = 0;
    YEARS_DIFF    INT = 0;
BEGIN
    IF UNITS IN ('YY', 'YYYY', 'YEAR', 'MM', 'M', 'MONTH') THEN
        YEARS_DIFF = DATE_PART('YEAR', END_T) - DATE_PART('YEAR', START_T);

        IF UNITS IN ('YY', 'YYYY', 'YEAR') THEN
            -- SQL SERVER DOES NOT COUNT FULL YEARS PASSED (ONLY DIFFERENCE BETWEEN YEAR PARTS)
            RETURN YEARS_DIFF;
        ELSE
            -- IF END MONTH IS LESS THAN START MONTH IT WILL SUBTRACTED
            RETURN YEARS_DIFF * 12 + (DATE_PART('MONTH', END_T) - DATE_PART('MONTH', START_T));
        END IF;
    END IF;

    -- MINUS OPERATOR RETURNS INTERVAL 'DDD DAYS HH:MI:SS'
    DIFF_INTERVAL = END_T - START_T;

    DIFF = DIFF + DATE_PART('DAY', DIFF_INTERVAL);

    IF UNITS IN ('WK', 'WW', 'WEEK') THEN
        DIFF = DIFF / 7;
        RETURN DIFF;
    END IF;

    IF UNITS IN ('DD', 'D', 'DAY') THEN
        RETURN DIFF;
    END IF;

    DIFF = DIFF * 24 + DATE_PART('HOUR', DIFF_INTERVAL);

    IF UNITS IN ('HH', 'HOUR') THEN
        RETURN DIFF;
    END IF;

    DIFF = DIFF * 60 + DATE_PART('MINUTE', DIFF_INTERVAL);

    IF UNITS IN ('MI', 'N', 'MINUTE') THEN
        RETURN DIFF;
    END IF;

    DIFF = DIFF * 60 + DATE_PART('SECOND', DIFF_INTERVAL);

    -- RETURN THE DIFFERENCE BETWEEN DATES TO THE SPECIFIED VALUE.
    RETURN DIFF;
END;
$$ LANGUAGE PLPGSQL;

-- Вспомогательная функция для поиска информации по клиентам.
-- Время ожидания, время обслуживания, последняя дата регистрации в очереди, последняя дата вызова клиента, последняя дата выхода из очереди.
CREATE OR REPLACE FUNCTION GET_CLIENT_SESSIONS()
    RETURNS TABLE
            (
                CLIENT_ID BIGINT,
                WAITING_TIME   INTEGER,
                SERVICE_TIME INTEGER,
                REG_DT    TIMESTAMP,
                CALL_DT   TIMESTAMP,
                FINISH_DT TIMESTAMP
            )
AS
$$
DECLARE
    EVENT_LOOP EVENT%ROWTYPE;
BEGIN
    -- СОЗДАТЬ ВРЕМЕННУЮ ТАБЛИЦУ ДЛЯ ВЫВОДА.
    CREATE TEMPORARY TABLE IF NOT EXISTS TEMP_CLIENT_SESSIONS
    (
        CLIENT_ID BIGINT,
        WAITING_TIME   INTEGER DEFAULT 0,
        SERVICE_TIME INTEGER DEFAULT 0,
        REG_DT    TIMESTAMP,
        CALL_DT   TIMESTAMP,
        FINISH_DT TIMESTAMP
    );

    -- СОЗДАТЬ ЦИКЛ ПО ТАБЛИЦЕ EVENT С КЛИЕНТСКИМИ ТИПАМИ СОБЫТИЙ.
    FOR EVENT_LOOP IN SELECT *
               FROM EVENT E
               WHERE E.CLIENT_ID IS NOT NULL
                 AND E.EVENT_TYPE_ID IN (9, 10, 11, 14, 15, 16, 19, 20, 21, 22, 23)
               ORDER BY CLIENT_ID, DATE_TIME_COLUMN
        LOOP
            -- РЕГИСТРАЦИЯ КЛИЕНТА [%S%S] НА УСЛУГУ [%S].
            IF EVENT_LOOP.EVENT_TYPE_ID = 9 THEN
                INSERT INTO TEMP_CLIENT_SESSIONS (CLIENT_ID, WAITING_TIME, SERVICE_TIME, REG_DT)
                VALUES (EVENT_LOOP.CLIENT_ID, 0, 0, EVENT_LOOP.DATE_TIME_COLUMN);
            END IF;

            -- РЕГИСТРАЦИЯ КЛИЕНТА [%S%S] НА УСЛУГУ  [%S] С ПОМОЩЬЮ СПЕЦИАЛЬНОЙ РЕГИСТРАЦИИ.
            IF EVENT_LOOP.EVENT_TYPE_ID = 14 THEN
                IF EXISTS(SELECT * FROM EVENT E WHERE EVENT_LOOP.CLIENT_ID = E.CLIENT_ID AND E.EVENT_TYPE_ID = 10) AND
                   NOT EXISTS(SELECT * FROM EVENT E WHERE EVENT_LOOP.CLIENT_ID = E.CLIENT_ID AND E.EVENT_TYPE_ID IN (15, 20))
                THEN
                    INSERT INTO TEMP_CLIENT_SESSIONS (CLIENT_ID, WAITING_TIME, SERVICE_TIME, REG_DT)
                    VALUES (EVENT_LOOP.CLIENT_ID, 0, 0, EVENT_LOOP.DATE_TIME_COLUMN);
                ELSE
                    INSERT INTO TEMP_CLIENT_SESSIONS (CLIENT_ID, WAITING_TIME, SERVICE_TIME, REG_DT, CALL_DT)
                    VALUES (EVENT_LOOP.CLIENT_ID, 0, 0, EVENT_LOOP.DATE_TIME_COLUMN, EVENT_LOOP.DATE_TIME_COLUMN);
                END IF;

            END IF;

            -- ВЫЗОВ КЛИЕНТА [%S%S] ПО УСЛУГЕ [%S].
            IF EVENT_LOOP.EVENT_TYPE_ID = 10 THEN
                UPDATE TEMP_CLIENT_SESSIONS TCS
                SET WAITING_TIME = TCS.WAITING_TIME +
                                   DATE_DIFF('SECOND', TCS.REG_DT::TIMESTAMP, EVENT_LOOP.DATE_TIME_COLUMN::TIMESTAMP),
                    CALL_DT = EVENT_LOOP.DATE_TIME_COLUMN
                WHERE EVENT_LOOP.CLIENT_ID = TCS.CLIENT_ID;
            END IF;

            -- ПЕРЕВОД КЛИЕНТА [%S%S] ИЗ ОКНА [%S] В ОКНО [%S]. ПРИЧИНА ПЕРЕВОДА: [%S].
            -- НАЗНАЧЕНИЕ КЛИЕНТА [%S%S] ОКНУ [%S]. ПРИЧИНА ПЕРЕВОДА: [%S].
            IF (EVENT_LOOP.EVENT_TYPE_ID IN (15, 23)) THEN
                UPDATE TEMP_CLIENT_SESSIONS TCS
                SET SERVICE_TIME = TCS.SERVICE_TIME +
                                DATE_DIFF('SECOND', TCS.CALL_DT::TIMESTAMP, EVENT_LOOP.DATE_TIME_COLUMN::TIMESTAMP),
                    REG_DT    = EVENT_LOOP.DATE_TIME_COLUMN
                WHERE EVENT_LOOP.CLIENT_ID = TCS.CLIENT_ID;
            END IF;

            -- ОТКЛАДЫВАНИЕ КЛИЕНТА [%S%S] НА [%D] МИНУТ.
            -- СБРОС КЛИЕНТА [%S%S].
            IF (EVENT_LOOP.EVENT_TYPE_ID IN (19, 22)) THEN
                UPDATE TEMP_CLIENT_SESSIONS TCS
                SET SERVICE_TIME = TCS.SERVICE_TIME +
                                DATE_DIFF('SECOND', TCS.CALL_DT::TIMESTAMP, EVENT_LOOP.DATE_TIME_COLUMN::TIMESTAMP)
                WHERE EVENT_LOOP.CLIENT_ID = TCS.CLIENT_ID;
            END IF;

            -- ВОЗВРАЩЕНИЕ КЛИЕНТА [%S%S] В ОЧЕРЕДЬ.
            IF (EVENT_LOOP.EVENT_TYPE_ID = 20) THEN
                UPDATE TEMP_CLIENT_SESSIONS TCS
                SET REG_DT = EVENT_LOOP.DATE_TIME_COLUMN
                WHERE EVENT_LOOP.CLIENT_ID = TCS.CLIENT_ID;
            END IF;

            -- ЗАВЕРШЕНИЕ ОБСЛУЖИВАНИЯ КЛИЕНТА [%S%S] ПО УСЛУГЕ [%S].
            IF (EVENT_LOOP.EVENT_TYPE_ID = 11) THEN
                UPDATE TEMP_CLIENT_SESSIONS TCS
                SET SERVICE_TIME = TCS.SERVICE_TIME +
                                DATE_DIFF('SECOND', TCS.CALL_DT::TIMESTAMP, EVENT_LOOP.DATE_TIME_COLUMN::TIMESTAMP),
                    FINISH_DT = EVENT_LOOP.DATE_TIME_COLUMN
                WHERE EVENT_LOOP.CLIENT_ID = TCS.CLIENT_ID;
            END IF;
        END LOOP;

    -- ВЕРНУТЬ РЕЗУЛЬТАТ РАБОТЫ ФУНКЦИИ ИЗ ВРЕМЕННОЙ ТАБЛИЦЫ.
    RETURN QUERY
        SELECT TCS.CLIENT_ID, TCS.WAITING_TIME, TCS.SERVICE_TIME, TCS.REG_DT, TCS.CALL_DT, TCS.FINISH_DT
        FROM TEMP_CLIENT_SESSIONS TCS;

    -- УДАЛИТЬ ВРЕМЕННУЮ ТАБЛИЦУ.
    DROP TABLE TEMP_CLIENT_SESSIONS;
END;
$$ LANGUAGE PLPGSQL;

-- Функция для отчета "Показатели обслуживания клиентов".
CREATE OR REPLACE FUNCTION REPORT_CLIENTS_SERVICE_METRICS(IN_DEPARTMENT_ID BIGINT,
                                                          IN_DATE_START TIMESTAMP,
                                                          IN_DATE_END TIMESTAMP,
                                                          IN_OPERATION_ID BIGINT,
                                                          IN_GROUP_QUALIFICATION_ID BIGINT,
                                                          IN_USER_ID BIGINT,
                                                          IN_CLIENT_WAITING_TIME INT,
                                                          IN_GROUP_MONTH BOOLEAN,
                                                          IN_GROUP_DAY BOOLEAN,
                                                          IN_GROUP_HOUR BOOLEAN,
                                                          IN_GROUP_GROUP_ID BOOLEAN,
                                                          IN_GROUP_CLIENT_SEGMENT_ID BOOLEAN,
                                                          IN_GROUP_DEPARTMENT_ID BOOLEAN,
                                                          IN_GROUP_USER_ID BOOLEAN,
                                                          IN_PREV_PAGE_LAST_ID BIGINT,
                                                          IN_PAGE_SIZE BIGINT,
                                                          IN_ORDER_BY_COLUMN_ID INT,
                                                          IN_IS_DESCENDING BOOLEAN,
                                                          IN_CURRENT_LANGUAGE_ID BIGINT)
    RETURNS TABLE
            (
                ID                               BIGINT,           -- Номер строки, ID
                DEPARTMENT_NAME                  TEXT,             -- Отделение
                OPERATION_GROUP_NAME             TEXT,             -- Группа
                USER_NAME                        TEXT,             -- ФИО ОПЕРАТОРА
                CLIENT_SEGMENT_NAME              TEXT,             -- СЕГМЕНТ КЛИЕНТОВ
                DATE_YEAR                        DOUBLE PRECISION, -- ГОД
                DATE_MONTH                       DOUBLE PRECISION, -- МЕСЯЦ
                DATE_DAY                         DOUBLE PRECISION, -- ДЕНЬ
                DATE_HOUR                        DOUBLE PRECISION, -- ЧАС
                NUMBER_TICKETS                   NUMERIC,          -- КОЛИЧЕСТВО ТАЛОНОВ
                TOTAL_NUMBER_OPERATIONS          NUMERIC,          -- ОБЩЕЕ КОЛИЧЕСТВО УСЛУГ
                NUMBER_SERVED_CLIENTS            NUMERIC,          -- ОБСЛУЖЕНЫЕ КЛИЕНТЫ, N
                PERCENT_SERVED_CLIENTS           NUMERIC,          -- ОБСЛУЖЕНЫЕ КЛИЕНТЫ, %
                NUMBER_CLIENTS_WAITED_NO_MOREN   NUMERIC,          -- КЛИЕНТЫ ОЖИДАВШИЕ НЕ БОЛЕЕ ЧЕМ N МИНУТ, N
                PERCENT_CLIENTS_WAITED_NO_MOREN  NUMERIC,          -- КЛИЕНТЫ ОЖИДАВШИЕ НЕ БОЛЕЕ ЧЕМ N МИНУТ, %
                NUMBER_OUT_OF_SERVICE_CUSTOMERS  NUMERIC,          -- КЛИЕНТЫ ОТКАЗАВШИЕСЯ ОТ ОБСЛУЖИВАНИЯ, N
                PERCENT_OUT_OF_SERVICE_CUSTOMERS NUMERIC,          -- КЛИЕНТЫ ОТКАЗАВШИЕСЯ ОТ ОБСЛУЖИВАНИЯ, %
                AVERAGE_SERVICE_TIME             NUMERIC,          -- СРЕДНЕЕ ВРЕМЯ ОБСЛУЖИВАНИЯ
                AVERAGE_WAITING_TIME             NUMERIC,          -- СРЕДНЕЕ ВРЕМЯ ОЖИДАНИЯ
                AVERAGE_TIME_SPENT_IN_DEPARTMENT NUMERIC           -- СРЕДНЕЕ ВРЕМЯ НАХОЖДЕНИЯ В ОТДЕЛЕНИИ
            )
AS
$$
BEGIN
    -- СОЗДАТЬ ВРЕМЕННУЮ ТАБЛИЦУ ДЛЯ ОБРАБОТКИ ДАННЫХ.
    CREATE TEMPORARY TABLE TEMP_RESULT_EVENTS
    (
        ID                               BIGINT,           -- НОМЕР СТРОКИ
        DEPARTMENT_ID                    BIGINT,           -- ОТДЕЛЕНИЕ, ID
        DEPARTMENT_NAME                  TEXT,             -- ОТДЕЛЕНИЕ, NAME
        GROUP_QUALIFICATION_ID           BIGINT,           -- ОПРЕДЕЛЕНИЕ ГРУППЫ, ID
        OPERATION_GROUP_NAME             TEXT,             -- ГРУППА, NAME
        USER_ID                          BIGINT,           -- ID ОПЕРАТОРА
        USER_NAME                        TEXT,             -- ФИО ОПЕРАТОРА
        CLIENT_SEGMENT_ID                BIGINT,           -- СЕГМЕНТ КЛИЕНТОВ, ID
        CLIENT_SEGMENT_NAME              TEXT,             -- СЕГМЕНТ КЛИЕНТОВ, NAME
        DATE_YEAR                        DOUBLE PRECISION, -- ГОД
        DATE_MONTH                       DOUBLE PRECISION, -- МЕСЯЦ
        DATE_DAY                         DOUBLE PRECISION, -- ДЕНЬ
        DATE_HOUR                        DOUBLE PRECISION, -- ЧАС
        NUMBER_TICKETS                   BIGINT,           -- КОЛИЧЕСТВО ТАЛОНОВ
        TOTAL_NUMBER_OPERATIONS          BIGINT,           -- ОБЩЕЕ КОЛИЧЕСТВО УСЛУГ
        NUMBER_SERVED_CLIENTS            BIGINT,           -- ОБСЛУЖЕНЫЕ КЛИЕНТЫ, N
        PERCENT_SERVED_CLIENTS           NUMERIC,          -- ОБСЛУЖЕНЫЕ КЛИЕНТЫ, %
        NUMBER_CLIENTS_WAITED_NO_MOREN   BIGINT,           -- КЛИЕНТЫ ОЖИДАВШИЕ НЕ БОЛЕЕ ЧЕМ N МИНУТ, N
        PERCENT_CLIENTS_WAITED_NO_MOREN  NUMERIC,          -- КЛИЕНТЫ ОЖИДАВШИЕ НЕ БОЛЕЕ ЧЕМ N МИНУТ, N
        NUMBER_OUT_OF_SERVICE_CUSTOMERS  BIGINT,           -- КЛИЕНТЫ ОТКАЗАВШИЕСЯ ОТ ОБСЛУЖИВАНИЯ, N
        PERCENT_OUT_OF_SERVICE_CUSTOMERS NUMERIC,          -- КЛИЕНТЫ ОТКАЗАВШИЕСЯ ОТ ОБСЛУЖИВАНИЯ, %
        AVERAGE_SERVICE_TIME             NUMERIC,          -- СРЕДНЕЕ ВРЕМЯ ОБСЛУЖИВАНИЯ
        AVERAGE_WAITING_TIME             NUMERIC,          -- СРЕДНЕЕ ВРЕМЯ ОЖИДАНИЯ
        AVERAGE_TIME_SPENT_IN_DEPARTMENT NUMERIC           -- СРЕДНЕЕ ВРЕМЯ НАХОЖДЕНИЯ В ОТДЕЛЕНИИ
    ) ON COMMIT DROP;

    -- СОЗДАТЬ ОТФИЛЬТРОВАННУЮ КОПИЮ ТАБЛИЦЫ EVENT.
    CREATE TEMPORARY TABLE POST_FILTERING_EVENTS ON COMMIT DROP AS
    SELECT *
    FROM EVENT E
    WHERE (E.DEPARTMENT_ID = IN_DEPARTMENT_ID OR IN_DEPARTMENT_ID = -1)
      AND E.DATE_TIME_COLUMN BETWEEN IN_DATE_START AND IN_DATE_END
      AND (E.OPERATION_ID = IN_OPERATION_ID OR IN_OPERATION_ID = -1)
      AND (E.GROUP_QUALIFICATION_ID = IN_GROUP_QUALIFICATION_ID OR IN_GROUP_QUALIFICATION_ID = -1)
      AND (E.USER_ID = IN_USER_ID OR IN_USER_ID = -1)
      AND E.CLIENT_ID IS NOT NULL;

    -- ОБРАБОТАТЬ ДАННЫЕ И ПОМЕСТИТЬ ВО ВРЕМЕННУЮ ТАБЛИЦУ.
    INSERT INTO TEMP_RESULT_EVENTS (ID, -- НОМЕР СТРОКИ
                                    DEPARTMENT_ID, -- ОТДЕЛЕНИЕ, ID
                                    OPERATION_GROUP_NAME, -- ГРУППА, ID
                                    USER_NAME, -- ФИО ОПЕРАТОРА
                                    CLIENT_SEGMENT_ID, -- СЕГМЕНТ КЛИЕНТОВ, ID
                                    DATE_YEAR, -- ГОД
                                    DATE_MONTH, -- МЕСЯЦ
                                    DATE_DAY, -- ДЕНЬ
                                    DATE_HOUR, -- ЧАС
                                    NUMBER_TICKETS, -- КОЛИЧЕСТВО ТАЛОНОВ
                                    TOTAL_NUMBER_OPERATIONS, -- ОБЩЕЕ КОЛИЧЕСТВО УСЛУГ
                                    NUMBER_SERVED_CLIENTS, -- ОБСЛУЖЕНЫЕ КЛИЕНТЫ, N
                                    NUMBER_CLIENTS_WAITED_NO_MOREN, -- КЛИЕНТЫ ОЖИДАВШИЕ НЕ БОЛЕЕ ЧЕМ N МИНУТ, N
                                    NUMBER_OUT_OF_SERVICE_CUSTOMERS, -- КЛИЕНТЫ ОТКАЗАВШИЕСЯ ОТ ОБСЛУЖИВАНИЯ, N
                                    AVERAGE_SERVICE_TIME, -- СРЕДНЕЕ ВРЕМЯ ОБСЛУЖИВАНИЯ
                                    AVERAGE_WAITING_TIME, -- СРЕДНЕЕ ВРЕМЯ ОЖИДАНИЯ
                                    AVERAGE_TIME_SPENT_IN_DEPARTMENT -- СРЕДНЕЕ ВРЕМЯ НАХОЖДЕНИЯ В ОТДЕЛЕНИИ
    )
    SELECT ROW_NUMBER() OVER (PARTITION BY 1)                                                                  ID,                              -- НОМЕР СТРОКИ
           CASE
               WHEN IN_GROUP_DEPARTMENT_ID = TRUE OR IN_GROUP_USER_ID = TRUE THEN PFE.DEPARTMENT_ID
               ELSE NULL END                                                                                   DEPARTMENT_ID,                   -- ОТДЕЛЕНИЕ ID
           CASE
               WHEN IN_GROUP_GROUP_ID = TRUE THEN
                   CASE
                       WHEN IN_CURRENT_LANGUAGE_ID = 41 -- АНГЛИЙСКИЙ
                           THEN PFE.OPERATION_GROUP_NAME_EN
                       WHEN IN_CURRENT_LANGUAGE_ID = 134 -- РУССКИЙ
                           THEN PFE.OPERATION_GROUP_NAME_RU
                       ELSE PFE.OPERATION_GROUP_NAME_CUSTOM_LANGUAGE -- КАСТОМНЫЙ ЯЗЫК
                       END
               ELSE NULL END                                                                                   OPERATION_GROUP_NAME,            -- ГРУППА NAME
           CASE
               WHEN IN_GROUP_USER_ID = TRUE THEN
                   CASE
                       WHEN PFE.USER_FIRST_NAME || ' ' || PFE.USER_MIDDLE_NAME || ' ' || PFE.USER_LAST_NAME IS NOT NULL THEN
                           PFE.USER_FIRST_NAME || ' ' || PFE.USER_MIDDLE_NAME || ' ' || PFE.USER_LAST_NAME
                       ELSE PFE.USER_LOGIN END
               ELSE NULL END                                                                                   USER_NAME,                       -- ФИО ОПЕРАТОРА
           CASE
               WHEN IN_GROUP_CLIENT_SEGMENT_ID = TRUE THEN PFE.CLIENT_SEGMENT_ID
               ELSE NULL END                                                                                   CLIENT_SEGMENT_ID,               -- СЕГМЕНТ КЛИЕНТОВ ID
           CASE
               WHEN IN_GROUP_MONTH = TRUE OR IN_GROUP_DAY = TRUE OR IN_GROUP_HOUR = TRUE
                   THEN DATE_PART('YEAR', PFE.DATE_TIME_COLUMN)
               ELSE NULL END                                                                                   DATE_YEAR,                       -- ГОД
           CASE
               WHEN IN_GROUP_MONTH = TRUE OR IN_GROUP_DAY = TRUE OR IN_GROUP_HOUR = TRUE
                   THEN DATE_PART('MONTH', PFE.DATE_TIME_COLUMN)
               ELSE NULL END                                                                                   DATE_MONTH,                      -- МЕСЯЦ
           CASE
               WHEN IN_GROUP_DAY = TRUE OR IN_GROUP_HOUR = TRUE THEN DATE_PART('DAY', PFE.DATE_TIME_COLUMN)
               ELSE NULL END                                                                                   DATE_DAY,                        -- ДЕНЬ
           CASE WHEN IN_GROUP_HOUR = TRUE THEN DATE_PART('HOUR', PFE.DATE_TIME_COLUMN) ELSE NULL END           DATE_HOUR,                       -- ЧАС
           COUNT(Q1.CLIENT_ID)                                                                                 NUMBER_TICKETS,                  -- КОЛИЧЕСТВО ТАЛОНОВ
           COUNT(1)                                                                                            TOTAL_NUMBER_OPERATIONS,         -- ОБЩЕЕ КОЛИЧЕСТВО УСЛУГ
           COUNT(Q2.CLIENT_ID)                                                                                 NUMBER_SERVED_CLIENTS,           -- ОБСЛУЖЕНЫЕ КЛИЕНТЫ, N
           COUNT(DISTINCT Q3.CLIENT_ID) FILTER (WHERE Q3.WAITING_TIME < IN_CLIENT_WAITING_TIME)                NUMBER_CLIENTS_WAITED_NO_MOREN,  -- КЛИЕНТЫ ОЖИДАВШИЕ НЕ БОЛЕЕ ЧЕМ N МИНУТ, N
           COUNT(DISTINCT PFE.CLIENT_ID) - COUNT(DISTINCT PFE.CLIENT_ID) FILTER (WHERE PFE.EVENT_TYPE_ID = 11) NUMBER_OUT_OF_SERVICE_CUSTOMERS, -- КЛИЕНТЫ ОТКАЗАВШИЕСЯ ОТ ОБСЛУЖИВАНИЯ, N
           AVG(DISTINCT Q3.SERVICE_TIME)                                                                       AVERAGE_SERVICE_TIME,            -- СРЕДНЕЕ ВРЕМЯ ОБСЛУЖИВАНИЯ
           AVG(DISTINCT Q3.WAITING_TIME)                                                                       AVERAGE_WAITING_TIME,            -- СРЕДНЕЕ ВРЕМЯ ОЖИДАНИЯ
           AVG(DISTINCT Q3.WAITING_TIME + Q3.SERVICE_TIME)                                                     AVERAGE_TIME_SPENT_IN_DEPARTMENT -- СРЕДНЕЕ ВРЕМЯ НАХОЖДЕНИЯ В ОТДЕЛЕНИИ
    FROM POST_FILTERING_EVENTS PFE
             LEFT JOIN (SELECT PFE1.CLIENT_ID, PFE1.EVENT_TYPE_ID
                        FROM POST_FILTERING_EVENTS PFE1
                        WHERE PFE1.EVENT_TYPE_ID IN (9, 14)) Q1
                       ON PFE.CLIENT_ID = Q1.CLIENT_ID AND PFE.EVENT_TYPE_ID = Q1.EVENT_TYPE_ID
             LEFT JOIN (SELECT PFE2.CLIENT_ID, PFE2.EVENT_TYPE_ID
                        FROM POST_FILTERING_EVENTS PFE2
                        WHERE PFE2.EVENT_TYPE_ID IN (11, 15, 19)) Q2
                       ON PFE.CLIENT_ID = Q2.CLIENT_ID AND PFE.EVENT_TYPE_ID = Q2.EVENT_TYPE_ID
             LEFT JOIN
             (SELECT * FROM GET_CLIENT_SESSIONS()) Q3 ON PFE.CLIENT_ID = Q3.CLIENT_ID
    GROUP BY CASE
                 WHEN IN_GROUP_MONTH = TRUE OR IN_GROUP_DAY = TRUE OR IN_GROUP_HOUR = TRUE
                     THEN DATE_PART('YEAR', PFE.DATE_TIME_COLUMN)
                 ELSE NULL END,
             CASE
                 WHEN IN_GROUP_MONTH = TRUE OR IN_GROUP_DAY = TRUE OR IN_GROUP_HOUR = TRUE
                     THEN DATE_PART('MONTH', PFE.DATE_TIME_COLUMN)
                 ELSE NULL END,
             CASE WHEN IN_GROUP_DAY = TRUE OR IN_GROUP_HOUR = TRUE THEN DATE_PART('DAY', PFE.DATE_TIME_COLUMN) ELSE NULL END,
             CASE WHEN IN_GROUP_HOUR = TRUE THEN DATE_PART('HOUR', PFE.DATE_TIME_COLUMN) ELSE NULL END,
             CASE
                 WHEN IN_GROUP_GROUP_ID = TRUE THEN
                     CASE
                         WHEN IN_CURRENT_LANGUAGE_ID = 41 -- АНГЛИЙСКИЙ
                             THEN PFE.OPERATION_GROUP_NAME_EN
                         WHEN IN_CURRENT_LANGUAGE_ID = 134 -- РУССКИЙ
                             THEN PFE.OPERATION_GROUP_NAME_RU
                         ELSE PFE.OPERATION_GROUP_NAME_CUSTOM_LANGUAGE -- КАСТОМНЫЙ ЯЗЫК
                         END
                 ELSE NULL END,
             CASE WHEN IN_GROUP_CLIENT_SEGMENT_ID = TRUE THEN PFE.CLIENT_SEGMENT_ID ELSE NULL END,
             CASE WHEN IN_GROUP_DEPARTMENT_ID = TRUE OR IN_GROUP_USER_ID = TRUE THEN PFE.DEPARTMENT_ID ELSE NULL END,
             CASE
                 WHEN IN_GROUP_USER_ID = TRUE THEN
                     CASE
                         WHEN PFE.USER_FIRST_NAME || ' ' || PFE.USER_MIDDLE_NAME || ' ' || PFE.USER_LAST_NAME IS NOT NULL THEN
                             PFE.USER_FIRST_NAME || ' ' || PFE.USER_MIDDLE_NAME || ' ' || PFE.USER_LAST_NAME
                         ELSE PFE.USER_LOGIN END
                 ELSE NULL END,
             CASE
                 WHEN NOT (IN_GROUP_MONTH = TRUE OR
                           IN_GROUP_DAY = TRUE OR
                           IN_GROUP_HOUR = TRUE OR
                           IN_GROUP_GROUP_ID = TRUE OR
                           IN_GROUP_CLIENT_SEGMENT_ID = TRUE OR
                           IN_GROUP_DEPARTMENT_ID = TRUE OR
                           IN_GROUP_USER_ID = TRUE) THEN PFE.ID
                 ELSE NULL END;

    -- СОЗДАТЬ ВРЕМЕННУЮ ТАБЛИЦУ TEMP_RESULT_TABLE ДЛЯ ВЫВОДА ДАННЫХ ИЗ ФУНКЦИИ.
    -- ДОБАВИТЬ СВЯЗЬ С ОТДЕЛЕНИЯМИ, СЕГМЕНТАМИ, ПРОЦЕНТНЫЕ ПОЛЯ.
    CREATE TEMPORARY TABLE TEMP_RESULT_TABLE ON COMMIT DROP AS
    SELECT TRE.ID                                                                                         ID,
           CASE
               WHEN TRE.DEPARTMENT_ID IS NOT NULL THEN
                   CASE
                       WHEN IN_CURRENT_LANGUAGE_ID = 41 -- АНГЛИЙСКИЙ
                           THEN (SELECT D.NAME_EN FROM DEPARTMENT D WHERE D.ID = TRE.DEPARTMENT_ID)
                       WHEN IN_CURRENT_LANGUAGE_ID = 134 -- РУССКИЙ
                           THEN (SELECT D.NAME_RU FROM DEPARTMENT D WHERE D.ID = TRE.DEPARTMENT_ID)
                       ELSE (SELECT D.NAME_CUSTOM_LANGUAGE FROM DEPARTMENT D WHERE D.ID = TRE.DEPARTMENT_ID) -- КАСТОМНЫЙ ЯЗЫК
                       END

               ELSE NULL END                                                                              DEPARTMENT_NAME,
           TRE.OPERATION_GROUP_NAME                                                                       OPERATION_GROUP_NAME,
           TRE.USER_NAME                                                                                  USER_NAME,
           CASE
               WHEN TRE.CLIENT_SEGMENT_ID IS NOT NULL THEN
                   CASE
                       WHEN IN_CURRENT_LANGUAGE_ID = 41 -- АНГЛИЙСКИЙ
                           THEN (SELECT C.NAME_EN FROM CLIENT_SEGMENT C WHERE C.ID = TRE.CLIENT_SEGMENT_ID)
                       WHEN IN_CURRENT_LANGUAGE_ID = 134 -- РУССКИЙ
                           THEN (SELECT C.NAME_RU FROM CLIENT_SEGMENT C WHERE C.ID = TRE.CLIENT_SEGMENT_ID)
                       ELSE (SELECT C.NAME_CUSTOM_LANGUAGE FROM CLIENT_SEGMENT C WHERE C.ID = TRE.CLIENT_SEGMENT_ID) -- КАСТОМНЫЙ ЯЗЫК
                       END

               ELSE NULL END                                                                              CLIENT_SEGMENT_NAME,
           TRE.DATE_YEAR                                                                                  DATE_YEAR,
           TRE.DATE_MONTH                                                                                 DATE_MONTH,
           TRE.DATE_DAY                                                                                   DATE_DAY,
           TRE.DATE_HOUR                                                                                  DATE_HOUR,
           TRE.NUMBER_TICKETS                                                                             NUMBER_TICKETS,
           TRE.TOTAL_NUMBER_OPERATIONS                                                                    TOTAL_NUMBER_OPERATIONS,
           TRE.NUMBER_SERVED_CLIENTS                                                                      NUMBER_SERVED_CLIENTS,
           (TRE.NUMBER_SERVED_CLIENTS / SUM(TRE.NUMBER_SERVED_CLIENTS) OVER ()) * 100                     PERCENT_SERVED_CLIENTS,           -- ОБСЛУЖЕНЫЕ КЛИЕНТЫ, %
           TRE.NUMBER_CLIENTS_WAITED_NO_MOREN                                                             NUMBER_CLIENTS_WAITED_NO_MOREN,
           (TRE.NUMBER_CLIENTS_WAITED_NO_MOREN / SUM(TRE.NUMBER_CLIENTS_WAITED_NO_MOREN) OVER ()) * 100   PERCENT_CLIENTS_WAITED_NO_MOREN,  -- КЛИЕНТЫ ОЖИДАВШИЕ НЕ БОЛЕЕ ЧЕМ N МИНУТ, %
           TRE.NUMBER_OUT_OF_SERVICE_CUSTOMERS                                                            NUMBER_OUT_OF_SERVICE_CUSTOMERS,
           (TRE.NUMBER_OUT_OF_SERVICE_CUSTOMERS / SUM(TRE.NUMBER_OUT_OF_SERVICE_CUSTOMERS) OVER ()) * 100 PERCENT_OUT_OF_SERVICE_CUSTOMERS, -- КЛИЕНТЫ ОТКАЗАВШИЕСЯ ОТ ОБСЛУЖИВАНИЯ, %
           TRE.AVERAGE_SERVICE_TIME                                                                       AVERAGE_SERVICE_TIME,
           TRE.AVERAGE_WAITING_TIME                                                                       AVERAGE_WAITING_TIME,
           TRE.AVERAGE_TIME_SPENT_IN_DEPARTMENT                                                           AVERAGE_TIME_SPENT_IN_DEPARTMENT
    FROM TEMP_RESULT_EVENTS TRE;

    -- ВЕРНУТЬ РЕЗУЛЬТАТ РАБОТЫ ФУНКЦИИ ИЗ ВРЕМЕННОЙ ТАБЛИЦЫ TEMP_RESULT_TABLE, ДОБАВИТЬ СОРТИРОВКУ, ПАГИНАЦИЮ И ИТОГИ.
    RETURN QUERY
        (SELECT *
         FROM TEMP_RESULT_TABLE TRT
         WHERE CASE WHEN IN_IS_DESCENDING = FALSE THEN TRT.ID > IN_PREV_PAGE_LAST_ID ELSE TRT.ID < IN_PREV_PAGE_LAST_ID END
         ORDER BY (CASE WHEN IN_ORDER_BY_COLUMN_ID = 1 AND IN_IS_DESCENDING = TRUE THEN TRT.ID END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 1 AND IN_IS_DESCENDING = FALSE THEN TRT.ID END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 2 AND IN_IS_DESCENDING = TRUE THEN TRT.DEPARTMENT_NAME END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 2 AND IN_IS_DESCENDING = FALSE THEN TRT.DEPARTMENT_NAME END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 3 AND IN_IS_DESCENDING = TRUE THEN TRT.OPERATION_GROUP_NAME END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 3 AND IN_IS_DESCENDING = FALSE THEN TRT.OPERATION_GROUP_NAME END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 4 AND IN_IS_DESCENDING = TRUE THEN TRT.USER_NAME END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 4 AND IN_IS_DESCENDING = FALSE THEN TRT.USER_NAME END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 5 AND IN_IS_DESCENDING = TRUE THEN TRT.CLIENT_SEGMENT_NAME END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 5 AND IN_IS_DESCENDING = FALSE THEN TRT.CLIENT_SEGMENT_NAME END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 6 AND IN_IS_DESCENDING = TRUE THEN TRT.DATE_YEAR END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 6 AND IN_IS_DESCENDING = FALSE THEN TRT.DATE_YEAR END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 7 AND IN_IS_DESCENDING = TRUE THEN TRT.DATE_MONTH END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 7 AND IN_IS_DESCENDING = FALSE THEN TRT.DATE_MONTH END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 8 AND IN_IS_DESCENDING = TRUE THEN TRT.DATE_DAY END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 8 AND IN_IS_DESCENDING = FALSE THEN TRT.DATE_DAY END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 9 AND IN_IS_DESCENDING = TRUE THEN TRT.DATE_HOUR END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 9 AND IN_IS_DESCENDING = FALSE THEN TRT.DATE_HOUR END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 10 AND IN_IS_DESCENDING = TRUE THEN TRT.NUMBER_TICKETS END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 10 AND IN_IS_DESCENDING = FALSE THEN TRT.NUMBER_TICKETS END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 11 AND IN_IS_DESCENDING = TRUE THEN TRT.TOTAL_NUMBER_OPERATIONS END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 11 AND IN_IS_DESCENDING = FALSE THEN TRT.TOTAL_NUMBER_OPERATIONS END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 12 AND IN_IS_DESCENDING = TRUE THEN TRT.NUMBER_SERVED_CLIENTS END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 12 AND IN_IS_DESCENDING = FALSE THEN TRT.NUMBER_SERVED_CLIENTS END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 13 AND IN_IS_DESCENDING = TRUE THEN TRT.PERCENT_SERVED_CLIENTS END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 13 AND IN_IS_DESCENDING = FALSE THEN TRT.PERCENT_SERVED_CLIENTS END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 14 AND IN_IS_DESCENDING = TRUE THEN TRT.NUMBER_CLIENTS_WAITED_NO_MOREN END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 14 AND IN_IS_DESCENDING = FALSE THEN TRT.NUMBER_CLIENTS_WAITED_NO_MOREN END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 15 AND IN_IS_DESCENDING = TRUE THEN TRT.PERCENT_CLIENTS_WAITED_NO_MOREN END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 15 AND IN_IS_DESCENDING = FALSE THEN TRT.PERCENT_CLIENTS_WAITED_NO_MOREN END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 16 AND IN_IS_DESCENDING = TRUE THEN TRT.NUMBER_OUT_OF_SERVICE_CUSTOMERS END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 16 AND IN_IS_DESCENDING = FALSE THEN TRT.NUMBER_OUT_OF_SERVICE_CUSTOMERS END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 17 AND IN_IS_DESCENDING = TRUE THEN TRT.PERCENT_OUT_OF_SERVICE_CUSTOMERS END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 17 AND IN_IS_DESCENDING = FALSE THEN TRT.PERCENT_OUT_OF_SERVICE_CUSTOMERS END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 18 AND IN_IS_DESCENDING = TRUE THEN TRT.AVERAGE_SERVICE_TIME END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 18 AND IN_IS_DESCENDING = FALSE THEN TRT.AVERAGE_SERVICE_TIME END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 19 AND IN_IS_DESCENDING = TRUE THEN TRT.AVERAGE_WAITING_TIME END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 19 AND IN_IS_DESCENDING = FALSE THEN TRT.AVERAGE_WAITING_TIME END) ASC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 20 AND IN_IS_DESCENDING = TRUE THEN TRT.AVERAGE_TIME_SPENT_IN_DEPARTMENT END) DESC,
                  (CASE WHEN IN_ORDER_BY_COLUMN_ID = 20 AND IN_IS_DESCENDING = FALSE THEN TRT.AVERAGE_TIME_SPENT_IN_DEPARTMENT END) ASC
         LIMIT CASE WHEN IN_PAGE_SIZE = -1 THEN NULL ELSE IN_PAGE_SIZE END
        )
        UNION ALL
        (SELECT MAX(TRT2.ID) + 1                           ID,                               -- НОМЕР СТРОКИ
                NULL                                       DEPARTMENT_NAME,                  -- ОТДЕЛЕНИЕ, NAME
                NULL                                       OPERATION_GROUP_NAME,             -- ГРУППА, NAME
                NULL                                       USER_NAME,                        -- ФИО ОПЕРАТОРА
                NULL                                       CLIENT_SEGMENT_NAME,              -- СЕГМЕНТ КЛИЕНТОВ, NAME
                NULL                                       DATE_YEAR,                        -- ГОД
                NULL                                       DATE_MONTH,                       -- МЕСЯЦ
                NULL                                       DATE_DAY,                         -- ДЕНЬ
                NULL                                       DATE_HOUR,                        -- ЧАС
                SUM(TRT2.NUMBER_TICKETS)                   NUMBER_TICKETS,                   -- КОЛИЧЕСТВО ТАЛОНОВ
                SUM(TRT2.TOTAL_NUMBER_OPERATIONS)          TOTAL_NUMBER_OPERATIONS,          -- ОБЩЕЕ КОЛИЧЕСТВО УСЛУГ
                SUM(TRT2.NUMBER_SERVED_CLIENTS)            NUMBER_SERVED_CLIENTS,            -- ОБСЛУЖЕНЫЕ КЛИЕНТЫ, N
                SUM(TRT2.PERCENT_SERVED_CLIENTS)           PERCENT_SERVED_CLIENTS,           -- ОБСЛУЖЕНЫЕ КЛИЕНТЫ, %
                SUM(TRT2.NUMBER_CLIENTS_WAITED_NO_MOREN)   NUMBER_CLIENTS_WAITED_NO_MOREN,   -- КЛИЕНТЫ ОЖИДАВШИЕ НЕ БОЛЕЕ ЧЕМ N МИНУТ, N
                SUM(TRT2.PERCENT_CLIENTS_WAITED_NO_MOREN)  PERCENT_CLIENTS_WAITED_NO_MOREN,  -- КЛИЕНТЫ ОЖИДАВШИЕ НЕ БОЛЕЕ ЧЕМ N МИНУТ, %
                SUM(TRT2.NUMBER_OUT_OF_SERVICE_CUSTOMERS)  NUMBER_OUT_OF_SERVICE_CUSTOMERS,  -- КЛИЕНТЫ ОТКАЗАВШИЕСЯ ОТ ОБСЛУЖИВАНИЯ, N
                SUM(TRT2.PERCENT_OUT_OF_SERVICE_CUSTOMERS) PERCENT_OUT_OF_SERVICE_CUSTOMERS, -- КЛИЕНТЫ ОТКАЗАВШИЕСЯ ОТ ОБСЛУЖИВАНИЯ, %
                AVG(TRT2.AVERAGE_SERVICE_TIME)             AVERAGE_SERVICE_TIME,             -- СРЕДНЕЕ ВРЕМЯ ОБСЛУЖИВАНИЯ
                AVG(TRT2.AVERAGE_WAITING_TIME)             AVERAGE_WAITING_TIME,             -- СРЕДНЕЕ ВРЕМЯ ОЖИДАНИЯ
                AVG(TRT2.AVERAGE_TIME_SPENT_IN_DEPARTMENT) AVERAGE_TIME_SPENT_IN_DEPARTMENT  -- СРЕДНЕЕ ВРЕМЯ НАХОЖДЕНИЯ В ОТДЕЛЕНИИ
         FROM TEMP_RESULT_TABLE TRT2);
END
$$ LANGUAGE PLPGSQL SECURITY DEFINER;